// SerialLink.cpp : Implementation of gdSerialLink

#include "stdafx.h"
#include "atlbase.h"
#include "atlsafe.h"
#include "SerialLink.h"
#include "gdSerialErrors.h"

#include "InternalEventSource.h"

extern CInternalEventSource G_EVENTSOURCE;

// CSerialLink

HRESULT CSerialLink::FinalConstruct()
{
	m_bPortOpen = FALSE;
	m_hCommFile = NULL;

	m_pPortName = new TCHAR [5];
	StrCpyN (m_pPortName, _T("COM1"), 5);

	m_dwPortSpeed = 2400;
	m_PortStopBits = ONESTOPBIT;

	m_bWaitCommEventPending = false;
	m_bSendPending = false;
	m_bReceivePending = false;

	m_nSendBufferSize = 0;
	m_pSendBuffer = NULL;

	m_nReceiveBufferSize = 0;
	m_pReceiveBuffer = NULL;

	CreateEvents ();

	return S_OK;
}

void CSerialLink::FinalRelease()
{
	if (m_bPortOpen) ClosePort ();

	DestroyEvents ();

	if (m_pPortName != NULL) delete m_pPortName;

	if (m_nSendBufferSize) delete m_pSendBuffer;
	if (m_nReceiveBufferSize) delete m_pReceiveBuffer;
}

HRESULT CSerialLink::CreateEvents (void)
{
	m_hEventCommEvent = ::CreateEvent (NULL, TRUE, FALSE, NULL);
	if (m_hEventCommEvent == NULL)
	{
		DWORD dwLastError = ::GetLastError ();
		ATLTRACE ("CreateEvent failed.  GetLastError returned 0x%08X\n", dwLastError);
		return E_FAIL;
	}

	m_hEventSend = ::CreateEvent (NULL, TRUE, FALSE, NULL);
	if (m_hEventSend == NULL)
	{
		DWORD dwLastError = ::GetLastError ();
		ATLTRACE ("CreateEvent failed.  GetLastError returned 0x%08X\n", dwLastError);
		return E_FAIL;
	}

	m_hEventReceive = ::CreateEvent (NULL, TRUE, FALSE, NULL);
	if (m_hEventReceive == NULL)
	{
		DWORD dwLastError = ::GetLastError ();
		ATLTRACE ("CreateEvent failed.  GetLastError returned 0x%08X\n", dwLastError);
		return E_FAIL;
	}

	G_EVENTSOURCE.AddPreEventWaitTarget (this);

	G_EVENTSOURCE.AddEvent (m_hEventCommEvent, this);
	G_EVENTSOURCE.AddEvent (m_hEventSend, this);
	G_EVENTSOURCE.AddEvent (m_hEventReceive, this);

	return S_OK;
}

HRESULT CSerialLink::DestroyEvents (void)
{
	G_EVENTSOURCE.RemoveEvent (m_hEventCommEvent);
	G_EVENTSOURCE.RemoveEvent (m_hEventSend);
	G_EVENTSOURCE.RemoveEvent (m_hEventReceive);

	G_EVENTSOURCE.RemovePreEventWaitTarget (this);

	if (FAILED (::CloseHandle (m_hEventCommEvent))) return E_FAIL;
	if (FAILED (::CloseHandle (m_hEventSend))) return E_FAIL;
	if (FAILED (::CloseHandle (m_hEventReceive))) return E_FAIL;

	return S_OK;
}

HRESULT CSerialLink::PreEventWait (void)
{
	ATLTRACE ("CSerialLink::OnPreEventWait\n");

	if (!m_bPortOpen) return S_OK;

	DWORD dwErrors;
	COMSTAT CommStatus;

	if (!::ClearCommError (m_hCommFile, &dwErrors, &CommStatus))
	{
		ATLTRACE ("ERROR: ClearCommError failed\n");
		return E_FAIL;
	}

	if (dwErrors) __raise OnError (static_cast <INT> (dwErrors));

	DWORD dwReceivedBytesInBuffer = CommStatus.cbInQue;

	if (dwReceivedBytesInBuffer)
	{
		if (!(m_bSendPending || m_bReceivePending))
		{
			if (FAILED (StartReadFile (dwReceivedBytesInBuffer)))
			{
				ATLTRACE ("ERROR: StartReadFile failed\n");
			}
		}
	}

	if (m_nSendBufferSize)
	{
		if (!(m_bSendPending || m_bReceivePending))
		{
			if (FAILED (StartWriteFile (m_pSendBuffer, m_nSendBufferSize)))
			{
				ATLTRACE ("ERROR: StartWriteFile failed\n");
			}

			m_nSendBufferSize = 0;
			delete m_pSendBuffer;
		}
	}

	if (!m_bWaitCommEventPending)
	{
		if (FAILED (StartWaitCommEvent ()))
		{
			ATLTRACE ("ERROR: StartWaitCommEvent failed\n");
		}
	}

	return S_OK;
}

HRESULT CSerialLink::OnEvent (HANDLE hEvent)
{
	ATLTRACE ("CSerialLink::OnEvent\n");

	if (!::ResetEvent (hEvent))
	{
		ATLTRACE ("WARNING: Couldn't reset event to nonsignaled state\n");
		return E_FAIL;
	}

	HRESULT hResult;

	if (!m_bPortOpen)
	{
		ATLTRACE ("OnEvent failed because the port was not open\n");
		hResult = E_PORT_NOT_OPENED;
	}
	else if (hEvent == m_hEventCommEvent)
	{
		ATLTRACE ("Received a CommEvent\n");
		hResult = OnCommEvent ();
		//StartWaitCommEvent ();
	}
	else if	(hEvent == m_hEventSend)
	{
		ATLTRACE ("Received a SendEvent\n");
		hResult = OnSendEvent ();
	}
	else if (hEvent == m_hEventReceive)
	{
		ATLTRACE ("Received a ReceiveEvent\n");
		hResult = OnReceiveEvent ();
		//StartReadFile (1);
	}
	else
	{
		ATLTRACE ("WARNING: OnEvent received an unexpected event\n");
		hResult = E_UNEXPECTED;
	}

	ATLTRACE ("OnEvent returning\n");

	return hResult;
}

STDMETHODIMP CSerialLink::get_PortName(BSTR* pVal)
{
	CComBSTR bstrPortName (m_pPortName);
	if (FAILED (bstrPortName.CopyTo (pVal))) return E_FAIL;

	return S_OK;
}

STDMETHODIMP CSerialLink::put_PortName(BSTR newVal)
{
	CComBSTR bstrPortName (newVal);
	int nPortNameLength = bstrPortName.Length () + 1; // length including null terminator

	LPTSTR pPortName = new TCHAR [nPortNameLength];
	if (pPortName == NULL) return E_OUTOFMEMORY;

	::StrCpyN (pPortName, CW2T (bstrPortName), nPortNameLength);

	if (m_pPortName != NULL) delete m_pPortName;

	m_pPortName = pPortName;

	return S_OK;
}

STDMETHODIMP CSerialLink::get_PortSpeed(LONG* pVal)
{
	*pVal = m_dwPortSpeed;

	return S_OK;
}

STDMETHODIMP CSerialLink::put_PortSpeed(LONG newVal)
{
	if ((newVal < 110) || (newVal > 256000)) return E_INVALIDARG;

	m_dwPortSpeed = newVal;

	return S_OK;
}

STDMETHODIMP CSerialLink::get_PortStopBits(FLOAT* pVal)
{
	switch (m_PortStopBits)
	{
		case (ONESTOPBIT):
			*pVal = 1;
			break;
		case (ONE5STOPBITS):
			*pVal = 1.5;
			break;
		case (TWOSTOPBITS):
			*pVal = 2;
			break;
		default:
			*pVal = 0;
	}

	return S_OK;
}

STDMETHODIMP CSerialLink::put_PortStopBits(FLOAT newVal)
{
	if (newVal == 1)
	{
		m_PortStopBits = ONESTOPBIT;
	}
	else if (newVal == 1.5)
	{
		m_PortStopBits = ONE5STOPBITS;
	}
	else if (newVal == 2)
	{
		m_PortStopBits = TWOSTOPBITS;
	}
	else
	{
		return E_INVALIDARG;
	}

	return S_OK;
}

STDMETHODIMP CSerialLink::OpenPort(void)
{
	ATLTRACE ("CSerialLink::OpenPort\n");

	if (m_bPortOpen) return E_PORT_PREVIOUSLY_OPENED;

	m_hCommFile = ::CreateFile (m_pPortName, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, FILE_FLAG_OVERLAPPED, NULL);
	if (m_hCommFile == INVALID_HANDLE_VALUE)
	{
		ATLTRACE ("ERROR: CreateFile returned INVALID_HANDLE_VALUE\n");
		return E_FAIL;
	}

	DWORD dwErrors;
	COMSTAT ComStatus;
	if (!::ClearCommError (m_hCommFile, &dwErrors, &ComStatus))
	{
		ATLTRACE ("ERROR: ClearCommError failed\n");
		return E_FAIL;
	}

	DCB dcb;
	if (!::GetCommState (m_hCommFile, &dcb))
	{
		ATLTRACE ("ERROR: GetCommState failed\n");
		return E_FAIL;
	}

	dcb.BaudRate = m_dwPortSpeed;
	dcb.fBinary = TRUE;
	dcb.fParity = FALSE;
	dcb.fOutxCtsFlow = FALSE;
	dcb.fOutxDsrFlow = FALSE;
	dcb.fDtrControl = DTR_CONTROL_DISABLE;
	//dcb.fDtrControl = DTR_CONTROL_ENABLE;
	//dcb.fDtrControl = DTR_CONTROL_HANDSHAKE;
	dcb.fDsrSensitivity = FALSE;
	dcb.fTXContinueOnXoff = FALSE;
	dcb.fOutX = FALSE;
	dcb.fInX = FALSE;
	dcb.fErrorChar = FALSE;
	dcb.fNull = FALSE;
	dcb.fRtsControl = RTS_CONTROL_DISABLE;
	//dcb.fRtsControl = RTS_CONTROL_HANDSHAKE;
	dcb.fAbortOnError = FALSE;
	dcb.wReserved = 0;
	dcb.ByteSize = 8;
	dcb.Parity = NOPARITY;
	dcb.StopBits = m_PortStopBits;

	if (!::SetCommState (m_hCommFile, &dcb))
	{
		ATLTRACE ("ERROR: SetCommState failed\n");
		return E_FAIL;
	}

	if (!SetupComm (m_hCommFile, 2048, 1024))
	{
		ATLTRACE ("ERROR: SetupComm failed\n");
		return E_FAIL;
	}

	COMMTIMEOUTS cto;
	cto.ReadIntervalTimeout = MAXDWORD;
	cto.ReadTotalTimeoutMultiplier = 0;
	cto.ReadTotalTimeoutConstant = 0;
	cto.WriteTotalTimeoutMultiplier = 0;
	cto.WriteTotalTimeoutConstant = 0;

	if (!::SetCommTimeouts (m_hCommFile, &cto))
	{
		ATLTRACE ("ERROR: SetCommTimeouts failed\n");
		return E_FAIL;
	}

	if (!::SetCommMask (m_hCommFile, EV_BREAK | EV_ERR | EV_RXCHAR))
	{
		ATLTRACE ("ERROR: SetCommMask failed\n");
		return E_FAIL;
	}

	m_bWaitCommEventPending = false;
	m_bSendPending = false;
	m_bReceivePending = false;

	if (m_nSendBufferSize) delete m_pSendBuffer;
	if (m_nReceiveBufferSize) delete m_pReceiveBuffer;

	m_bPortOpen = TRUE;

	//HRESULT hResult;

	//hResult = StartWaitCommEvent ();
	//if (FAILED (hResult)) return hResult;

	//hResult = StartReadFile (1);
	//if (FAILED (hResult)) return hResult;

	ATLTRACE ("CSerialLink::OpenPort returning successfully\n");
	return S_OK;
}

STDMETHODIMP CSerialLink::ClosePort(void)
{
	if (!m_bPortOpen) return E_PORT_NOT_OPENED;

	if (!::CloseHandle (m_hCommFile)) return E_FAIL;

	m_bPortOpen = FALSE;

	return S_OK;
}

STDMETHODIMP CSerialLink::SendData(BYTE *Data, LONG DataLength)
{
	if (!m_bPortOpen) return E_PORT_NOT_OPENED;

	if (m_nSendBufferSize)
	{
		int nNewSendBufferSize = m_nSendBufferSize + DataLength;
		BYTE *pNewSendBuffer = new BYTE [nNewSendBufferSize];

		memcpy (pNewSendBuffer, m_pSendBuffer, m_nSendBufferSize);
		memcpy (pNewSendBuffer + m_nSendBufferSize, Data, DataLength);

		delete m_pSendBuffer;

		m_nSendBufferSize = nNewSendBufferSize;
		m_pSendBuffer = pNewSendBuffer;
	}
	else
	{
		m_nSendBufferSize = DataLength;
		m_pSendBuffer = new BYTE [DataLength];

		memcpy (m_pSendBuffer, Data, DataLength);
	}

	return S_OK;
}

STDMETHODIMP CSerialLink::SendDataArray(SAFEARRAY* Data)
{
	// create a CComSafeArray object and attach it to the Data SAFEARRAY
	CComSafeArray <BYTE> DataArray (Data);

	// make sure the array is an array of BYTEs
	VARTYPE VariableType = DataArray.GetType ();
	if (VariableType != VT_I1 && VariableType != VT_UI1) return E_INVALIDARG;

	// make sure the array is only one dimension
	if (DataArray.GetDimensions () != 1) return E_INVALIDARG;

	// get the length of the data
	int nDataLength = DataArray.GetCount ();

	// get a pointer to the data
	BYTE *pData;
	if (FAILED (::SafeArrayAccessData (DataArray.m_psa, (LPVOID *) &pData))) return E_FAIL;

	// send the data
	if (FAILED (SendData (pData, nDataLength))) return E_FAIL;

	// release the pointer to the data
	if (FAILED (::SafeArrayUnaccessData (DataArray.m_psa))) return E_FAIL;

	return S_OK;
}

STDMETHODIMP CSerialLink::SendDataString(BSTR Data)
{
	CComBSTR bstrData (Data);
	int nLength = bstrData.Length (); // remove trailing null

	CW2A W2AConverter (bstrData);
	BYTE *pData = (BYTE *) ((LPSTR) W2AConverter);

	return SendData (pData, nLength);
}

HRESULT CSerialLink::StartWaitCommEvent (void)
{
	ATLTRACE ("CSerialLink::StartWaitCommEvent\n");

	if (!m_bPortOpen) return E_PORT_NOT_OPENED;

	m_OverlappedCommEvent.Internal = 0;
	m_OverlappedCommEvent.InternalHigh = 0;
	m_OverlappedCommEvent.Offset = 0;
	m_OverlappedCommEvent.OffsetHigh = 0;
	m_OverlappedCommEvent.hEvent = m_hEventCommEvent;

	if (!::WaitCommEvent (m_hCommFile, &m_dwCommEvent, &m_OverlappedCommEvent))
	{
		if (::GetLastError () == ERROR_IO_PENDING)
		{
			m_bWaitCommEventPending = true;
		}
		else
		{
			ATLTRACE ("ERROR: WaitCommEvent failed\n");
			return E_FAIL;
		}
	}
	else
	{
		ATLTRACE ("WaitCommEvent returned synchronously\n");
	}

	ATLTRACE ("StartWaitCommEvent returning successfully\n");
	return S_OK;
}

HRESULT CSerialLink::StartWriteFile (BYTE *Data, LONG DataLength)
{
	if (!m_bPortOpen) return E_PORT_NOT_OPENED;

	m_OverlappedSend.Internal = 0;
	m_OverlappedSend.InternalHigh = DataLength;
	m_OverlappedSend.Offset = 0;
	m_OverlappedSend.OffsetHigh = 0;
	m_OverlappedSend.hEvent = m_hEventSend;

	DWORD dwBytesSent;
	if (!::WriteFile (m_hCommFile, Data, DataLength, &dwBytesSent, &m_OverlappedSend))
	{
		if (::GetLastError () == ERROR_IO_PENDING)
		{
			m_bSendPending = true;
		}
		else
		{
			ATLTRACE ("ERROR: WriteFile failed\n");
			return E_FAIL;
		}
	}
	else
	{
		ATLTRACE ("WriteFile returned synchronously\n");
	}

	return S_OK;
}

HRESULT CSerialLink::StartReadFile (DWORD dwReceivedBytesInBuffer)
{
	ATLTRACE ("CSerialLink::StartReadFile\n");

	if (!m_bPortOpen) return E_PORT_NOT_OPENED;

	/*
	DWORD dwError;
	COMSTAT ComStatus;
	if (!::ClearCommError (m_hCommFile, &dwError, &ComStatus))
	{
		ATLTRACE ("ERROR: ClearCommError failed\n");
		return E_FAIL;
	}

	if (dwError != 0)
	{
		ATLTRACE ("WARNING: A communications error has occurred\n");
		__raise OnError (static_cast <INT> (dwError));
	}
	*/

	//DWORD dwBytesInBuffer = ComStatus.cbInQue;
	//if (dwBytesInBuffer == 0) dwBytesInBuffer = 1;
	//DWORD dwBytesInBuffer = 512;

	// delete the old receive buffer
	if (m_pReceiveBuffer != NULL) delete m_pReceiveBuffer;

	// create a new receive buffer
	m_nReceiveBufferSize = static_cast <int> (dwReceivedBytesInBuffer);
	m_pReceiveBuffer = new BYTE [dwReceivedBytesInBuffer];

	m_OverlappedReceive.Internal = 0;
	m_OverlappedReceive.InternalHigh = dwReceivedBytesInBuffer;
	m_OverlappedReceive.Offset = 0;
	m_OverlappedReceive.OffsetHigh = 0;
	m_OverlappedReceive.hEvent = m_hEventReceive;

	DWORD dwBytesReceived;
	if (!::ReadFile (m_hCommFile, m_pReceiveBuffer, m_nReceiveBufferSize, &dwBytesReceived, &m_OverlappedReceive))
	{
		if (::GetLastError () == ERROR_IO_PENDING)
		{
			m_bReceivePending = true;
		}
		else
		{
			ATLTRACE ("ERROR: ReadFile failed\n");
			return E_FAIL;
		}
	}
	else
	{
		ATLTRACE ("ReadFile returned synchronously\n");
		ATLTRACE ("ReadFile read %d bytes\n", dwBytesReceived);
	}
	
	ATLTRACE ("CSerialLink::StartReadFile returning successfully\n");
	return S_OK;
}

HRESULT CSerialLink::OnCommEvent (void)
{
	ATLTRACE ("CSerialLink::OnCommEvent\n");

	m_bWaitCommEventPending = false;

	/*
	DWORD dwDummy;
	if (!::GetOverlappedResult (m_hCommFile, &m_OverlappedCommEvent, &dwDummy, FALSE))
	{
		ATLTRACE ("ERROR: GetOverlappedResult failed\n");
		return HRESULT_FROM_WIN32 (::GetLastError ());
	}

	DWORD dwError;
	COMSTAT ComStatus;
	if (!::ClearCommError (m_hCommFile, &dwError, &ComStatus))
	{
		ATLTRACE ("ERROR: ClearCommError failed\n");
		return E_FAIL;
	}

	if (dwError != 0)
	{
		ATLTRACE ("WARNING: A communications error has occurred\n");
		__raise OnError (static_cast <INT> (dwError));
	}

	if (m_dwCommEvent & EV_BREAK)
	{
		ATLTRACE ("OnCommEvent received an EV_BREAK event\n");
		if (!::ClearCommBreak (m_hCommFile))
		{
			ATLTRACE ("ERROR: ClearCommBreak failed\n");
			return E_FAIL;
		}
	}

	if (m_dwCommEvent & EV_RXCHAR)
	{
		ATLTRACE ("OnCommEvent received an EV_RXCHAR event\n");
		if (ComStatus.cbInQue > 0)
		{
			HRESULT hResult = StartReadFile (ComStatus.cbInQue);
			if (FAILED (hResult))
			{
				ATLTRACE ("ERROR: StartReadFile failed\n");
				return hResult;
			}
		}
	}

	ATLTRACE ("OnCommEvent returning successfully\n");
	*/
	return S_OK;
}

HRESULT CSerialLink::OnSendEvent (void)
{
	m_bSendPending = false;

	DWORD dwBytesSent;
	if (!::GetOverlappedResult (m_hCommFile, &m_OverlappedSend, &dwBytesSent, FALSE))
	{
		ATLTRACE ("ERROR: GetOverlappedResult failed\n");
		return HRESULT_FROM_WIN32 (::GetLastError ());
	}

	__raise OnSendCompleted (static_cast <LONG> (dwBytesSent));

	return S_OK;
}

HRESULT CSerialLink::OnReceiveEvent (void)
{
	ATLTRACE ("CSerialLink::OnReceiveEvent\n");

	m_bReceivePending = false;

	// get the actual number of bytes received
	DWORD dwBytesReceived;
	if (!::GetOverlappedResult (m_hCommFile, &m_OverlappedReceive, &dwBytesReceived, FALSE))
	{
		ATLTRACE ("ERROR: GetOverlappedResult failed\n");
		return HRESULT_FROM_WIN32 (::GetLastError ());
	}

	ATLTRACE ("GetOverlappedResult says %d bytes were received\n", dwBytesReceived);

	// we don't need to do anything if nothing was received
	if (dwBytesReceived == 0) return S_OK;

	// make sure read file didn't overrun our buffer for some reason... this would indicate a serious problem
	if (static_cast <int> (dwBytesReceived) > m_nReceiveBufferSize)
	{
		ATLTRACE ("ERROR: ReadFile read more bytes than the size of its buffer\n");
		return E_FAIL;
	}

	// create a safearray
	CComSafeArray <BYTE> ReceivedDataArray;
	if (FAILED (ReceivedDataArray.Create (dwBytesReceived))) return E_FAIL;

	// copy everything over to the safearray
	for (DWORD x = 0; x < dwBytesReceived; x++)
	{
		if (FAILED (ReceivedDataArray.SetAt (x, m_pReceiveBuffer [x]))) return E_FAIL;
	}
	
	// convert the ANSI buffer to a Unicode BSTR
	CComBSTR ReceivedDataString (dwBytesReceived, (LPCWSTR) CA2W ((LPCSTR) m_pReceiveBuffer));

	// raise the events
	__raise OnReceiveData (m_pReceiveBuffer, dwBytesReceived);
	__raise OnReceiveDataArray (ReceivedDataArray);
	__raise OnReceiveDataString (ReceivedDataString);

	return S_OK;
}
